import os
import tensorflow as tf

from abc import ABC, abstractmethod


class BaseNeuralNetwork(ABC):
    def __init__(self, save_dir='results'):
        self.save_dir = save_dir
        self.save_path = os.path.join(self.save_dir, 'best.ckpt')
        self.session = None
        self.x_t = None
        self.y_t = None
        self._logits = None
        self.init = None

    def init_model(self, X, y):
        self.session = tf.Session(config=tf.ConfigProto(log_device_placement=True))
        self.prediction
        self.optimize
        self.error
        self.init = tf.global_variables_initializer()
        self.session.run(self.init)
        self.additional_inits()

    @abstractmethod
    def save(self, save_path):
        pass

    @abstractmethod
    def restore(self, save_path):
        pass

    def additional_inits(self):
        pass

    @property
    @abstractmethod
    def prediction(self):
        pass

    @property
    @abstractmethod
    def loss(self):
        pass

    @property
    @abstractmethod
    def optimize(self):
        pass

    @property
    @abstractmethod
    def error(self):
        pass



from sklearn.exceptions import NotFittedError
from sklearn.naive_bayes import MultinomialNB
from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.model_selection import GridSearchCV

from pylat.wrapper.transformer.text_preprocessor import TextPreprocessor
from pylat.exceptions import InvalidArgumentError

import logging
import pytest
import unittest

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def pipe_sample(doc):
    if len(doc) > 10:
        logger.info('This is a pretty long text!')
    else:
        logger.info('This is a pretty short text!')
    return doc


def do_nothing(x):
    return x


class TestTextPreprocessor(unittest.TestCase):

    def setUp(self):
        self.sample_en = [
            u"Hello, this is a simple text case :)",
            u"This is the 2nd sample",
            u"I would like to present you my brother-in-law",
            u"This is pretty sad",
            u"I need to make multiple samples",
            u"For the scikit learn grid search",
            u"I'm tired :("
        ]
        self.labels_en = [
            0,
            1,
            0,
            1,
            1,
            0,
            1
        ]

    def test_invalid_params(self):
        invalid_stop_word_type = TextPreprocessor(remove_stop_words=2)
        with pytest.raises(InvalidArgumentError):
            invalid_stop_word_type.fit(self.sample_en)

        invalid_lemma_type = TextPreprocessor(lemmatize=3)
        with pytest.raises(InvalidArgumentError):
            invalid_lemma_type.fit(self.sample_en)

        invalid_add_pipe_type = TextPreprocessor(additional_pipes='hi')
        with pytest.raises(InvalidArgumentError):
            invalid_add_pipe_type.fit(self.sample_en)

        invalid_model_id = TextPreprocessor(spacy_model_id='invented')
        with pytest.raises(InvalidArgumentError):
            invalid_model_id.fit(self.sample_en)

    def test_valid_params(self):
        text_processor = TextPreprocessor(spacy_model_id='es', remove_stop_words=True,
                                          lemmatize=True, additional_pipes=[pipe_sample])
        text_processor.fit(self.sample_en)
        transformed_text = text_processor.transform(self.sample_en)
        self.assertEqual(len(self.sample_en), len(transformed_text))
        for doc in transformed_text:
            self.assertTrue(len(doc) != 0)

    def test_sklearn_pipeline(self):
        text_preprocessor = TextPreprocessor(spacy_model_id='en', lemmatize=True)
        tf_idf_vectorizer = TfidfVectorizer(input='content', tokenizer=lambda x: x,
                                            preprocessor=None, lowercase=False)
        bayes_clf = MultinomialNB()

        pipe = Pipeline(steps=[
            ('preprocessing', text_preprocessor),
            ('tf-idf', tf_idf_vectorizer),
            ('clf', bayes_clf)
        ])
        pipe.fit(self.sample_en, self.labels_en)

    def test_search_cv(self):
        pipe = Pipeline(steps=[
            ('txt_prep', TextPreprocessor()),
            ('tf-idf', TfidfVectorizer(input='content', tokenizer=do_nothing,
                                       preprocessor=None, lowercase=False)),
            ('clf', MultinomialNB())
        ])

        param_grid = {
            'txt_prep__spacy_model_id': ['en', 'es'],
            'txt_prep__lemmatize': [True, False],
            'txt_prep__remove_stop_words': [True, False],
            'txt_prep__additional_pipes': [None, [pipe_sample]]
        }

        grid_search = GridSearchCV(pipe, param_grid, n_jobs=-1, iid=False, cv=2)
        grid_search.fit(self.sample_en, self.labels_en)

    def test_transform_before_fit(self):
        preprocessor = TextPreprocessor()
        with pytest.raises(NotFittedError):
            preprocessor.transform(self.sample_en)


if __name__ == '__main__':
    unittest.main()



from pylat.exceptions import InvalidArgumentError

import csv


class LexiconManager():
    """ This class produces new features from a text using a Lexicon.

    The LexiconManager class uses the NRC Valence, Arousal and Dominance
    lexicon and provides function the return a score for each of the three
    dimensions given a piece of text. This class can be used for NLP
    applications in order to add new features to a classifier.

    Parameters
    ----------
    lexicon_path: Path to the lexicon text file.

    Examples
    --------
    >>> import os
    >>> from pylat.wrapper.transformer.lexicon_manager import LexiconManager
    >>> tokens = ['I', 'need', 'more', 'alcohol', 'now']
    >>> lexicon_path = os.path.join('test', 'data', 'test_lexicon.txt')
    >>> lexicon = LexiconManager(lexicon_path)
    >>> round(lexicon.valence(tokens), 3)
        0.178
    >>> round(lexicon.arousal(tokens), 3)
        0.188
    >>> round(lexicon.dominance(tokens), 3)
        0.304
    """
    def __init__(self, lexicon_path):
        self._lexicon = dict()
        with open(lexicon_path, newline='') as csvfile:
            lexicon_reader = csv.DictReader(csvfile, delimiter='\t')
            for row in lexicon_reader:
                name = row['Word']
                valence = row['Valence']
                arousal = row['Arousal']
                dominance = row['Dominance']
                self._lexicon[name] = LexiconRow(valence, arousal, dominance)

    def valence(self, text):
        """ Returns the valence score for the given text.

        The valence score is a float between 0 and 1 that represents
        the pleasure--displeasure dimension of the text. A value near
        0 means that the text represents displeasure, while a value near
        1 means that the text represents pleasure.

        :param text: Iterable of tokens present in the text.
        :return: float normalized number with the valence score.
        """
        return self._obtain_lexicon_info('valence', text)

    def arousal(self, text):
        """ Returns the arousal score for the given text.

        The arousal score is a float between 0 and 1 that represents
        the active--passive dimension of the text. A value

        :param text: String representation of the text.
        :return: float normalized number with the arousal score.
        """
        return self._obtain_lexicon_info('arousal', text)

    def dominance(self, text):
        """ Returns the dominance score for the given text.

        :param text: String representation of the text
        :return: float normalized number with the dominance score
        """
        return self._obtain_lexicon_info('dominance', text)

    def _obtain_lexicon_info(self, variable, text):
        """ Calculates a specific dimension from the lexicon.

        :param variable: Must be one of 'valence', 'arousal' and 'dominance'.
        :param text: List of tokens which are present in the text.
        :return: float normalized number with the specified score
        """
        if not hasattr(text, '__iter__') or type(text) == str:
            raise InvalidArgumentError('text', 'Text must be an iterable of tokens')
        result = 0
        words_used = 0
        for token in text:
            lexicon_data = self._lexicon.get(token)
            if lexicon_data is not None:
                words_used += 1
                result += float(getattr(lexicon_data, variable))
        return 0.5 if words_used == 0 else result / words_used


class LexiconRow():
    def __init__(self, valence, arousal, dominance):
        self.valence = valence
        self.arousal = arousal
        self.dominance = dominance



class DataSample():
    def __init__(self, msg_id, msg_text, board_id,
                 root_id, kudos, author_id, post_time,
                 edit_time, thread_id, parent_msg_id,
                 views, label):
        self.msg_id = msg_id
        self.msg_text = msg_text
        self.board_id = board_id
        self.root_id = root_id
        self.kudos = kudos
        self.author_id = author_id
        self.post_time = post_time
        self.edit_time = edit_time
        self.thread_id = thread_id
        self.parent_msg_id = parent_msg_id
        self.views = views
        self.label = label

    def to_df_data(self):
        return {
            'id': self.msg_id,
            'text': self.msg_text,
            'board': self.board_id,
            'root': self.root_id,
            'kudos': self.kudos,
            'author_id': self.author_id,
            'post_time': self.post_time,
            'edit_time': self.edit_time,
            'thread': self.thread_id,
            'parent': self.parent_msg_id,
            'views': self.views,
            'label': self.label
        }



import pandas as pd
import xmltodict

import csv
import logging
import os

from .data_sample import DataSample


class DataParser:
    def __init__(self, data_path, labels_file_name, posts_dir_name):
        self.labels_file = os.path.join(data_path, labels_file_name)
        self.posts_dir = os.path.join(data_path, posts_dir_name)
        self._labeled_dataframe = None
        self._complete_dataframe = None

    @property
    def labeled_dataframe(self):
        if self._labeled_dataframe is not None:
            return self._labeled_dataframe
        else:
            self._labeled_dataframe = self._compute_dataframe(labeled_only=True)
            return self.labeled_dataframe

    @property
    def complete_dataframe(self):
        if self._complete_dataframe is not None:
            return self._complete_dataframe
        else:
            self._complete_dataframe = self._compute_dataframe()
            return self._complete_dataframe

    def _compute_dataframe(self, labeled_only=False):
        logging.info('Parsing labels...')
        self.labels = self._parse_labels()
        logging.info('Parsing posts...')
        if labeled_only:
            posts = [self._parse_sample(post) for post in os.listdir(self.posts_dir)
                     if self._is_labeled(post)]
        else:
            posts = [self._parse_sample(post) for post in os.listdir(self.posts_dir)]
        logging.info('Creating dataframe...')
        df = self._construct_df_from(posts)
        return df

    def _is_labeled(self, post):
        for label in self.labels.keys():
            if '-{}.xml'.format(label) in post:
                return True
        return False

    def _parse_sample(self, xml_doc):
        with open(os.path.join(self.posts_dir, xml_doc), 'r', encoding='utf8') as f:
            xml_data = f.read()
        tree = xmltodict.parse(xml_data)
        message = tree['response']['message']
        msg_id = int(message['id']['#text'])
        board_id = int(message['board_id']['#text'])
        root_id = DataParser._extract_id_from(message['root']['@href'])
        kudos = int(message['kudos']['count']['#text'])
        post_time = message['post_time']['#text']
        edit_time = message['last_edit_time']['#text']
        msg_data = DataParser._get_text(message)
        thread_id = DataParser._extract_id_from(message['thread']['@href'])
        parent_msg_id = DataParser._extract_parent_id_from(message)
        views = message['views']['count']['#text']
        author_id = DataParser._extract_id_from(message['author']['@href'])
        label = self.labels.get(msg_id)
        return DataSample(msg_id, msg_data, board_id, root_id,
                          kudos, author_id, post_time, edit_time,
                          thread_id, parent_msg_id, views, label)

    def _parse_labels(self):
        with open(self.labels_file) as csv_file:
            reader = csv.reader(csv_file, delimiter='\t')
            logging.debug('Reader created')
            labels = {int(row[0]): row[1] for row in reader}
            logging.debug('Labels obtained')
        return labels

    def _construct_df_from(self, posts):
        df = pd.DataFrame.from_dict([post.to_df_data() for post in posts])
        return df

    @classmethod
    def _get_text(cls, message):
        try:
            return message['body']['#text']
        except KeyError:
            return ''

    @classmethod
    def _extract_id_from(cls, href):
        return href.split('/')[-1]

    @classmethod
    def _extract_parent_id_from(cls, message):
        try:
            parent_id = message['parent']['@href']
            return parent_id
        except KeyError:
            return None



from pylat.wrapper.transformer.lexicon_manager import LexiconManager
from pylat.exceptions import InvalidArgumentError

import os
import pytest

lexicon_path = os.path.join('test', 'data', 'test_lexicon.txt')


class TestLexiconManager():
    def test_valid_text(self):
        """ Test with tokens that are contained in the lexicon.
        """
        tokens = ["this", "first", "test", "is", "mine"]
        lexicon = LexiconManager(lexicon_path)
        assert pytest.approx(lexicon.valence(tokens), 0.001) == 0.270
        assert pytest.approx(lexicon.arousal(tokens), 0.001) == 0.220
        assert pytest.approx(lexicon.dominance(tokens), 0.001) == 0.120

    def test_invalid_lexicon_path(self):
        """ Test passing an invalid lexicon path in the constructor.
        """
        with pytest.raises(FileNotFoundError):
            LexiconManager('invented')

    def test_invalid_text(self):
        """ Test with invalid token format.
        """
        text = "This is an untokenized text"
        lexicon = LexiconManager(lexicon_path)
        with pytest.raises(InvalidArgumentError):
            lexicon.valence(text)

    def test_text_without_info(self):
        """ Test where there is no token contained in the lexicon.
        """
        tokens = ['a', 'b', 'c']
        lexicon = LexiconManager(lexicon_path)
        assert pytest.approx(lexicon.valence(tokens), 0.1) == 0.5
        assert pytest.approx(lexicon.arousal(tokens), 0.1) == 0.5
        assert pytest.approx(lexicon.dominance(tokens), 0.1) == 0.5



from abc import ABC, abstractmethod
import tensorflow as tf

__author__ = 'Alejandro González Hevia'


class BaseCellFactory(ABC):
    """Abstract base class for all cell factories."""

    @abstractmethod
    def __call__(self, num_units, activation, kernel_init, dropout, layer_norm):
        pass


class LSTMCellFactory(BaseCellFactory):
    """Factory that creates variations of the LSTM cell."""

    def __call__(self, num_units, activation, kernel_init, dropout, layer_norm):
        if layer_norm:
            return tf.contrib.rnn.LayerNormBasicLSTMCell(num_units,
                                                         layer_norm=True)
        else:
            return tf.keras.layers.LSTMCell(num_units, dropout=dropout)


class GRUCellFactory(BaseCellFactory):
    """Factory that creates variations of the GRU cell."""

    def __call__(self, num_units, activation, kernel_init, dropout, layer_norm):
        return tf.keras.layers.GRUCell(num_units, activation,
                                       kernel_initializer=kernel_init,
                                       dropout=dropout)


class SimpleCellFactory(BaseCellFactory):
    """Factory that creates variations of a basic rnn cell."""

    def __call__(self, num_units, activation, kernel_init, dropout, layer_norm):
        return tf.keras.layers.SimpleRNNCell(num_units, activation,
                                             kernel_initializer=kernel_init,
                                             dropout=dropout)



from pylat.exceptions import InvalidArgumentError
from sklearn.base import BaseEstimator, TransformerMixin

import pandas as pd


class DataFrameSelector(BaseEstimator, TransformerMixin):
    """Transformer that receives a pandas Dataframe and returns a subset of its values as a numpy array.

    This transformer is meant to be used as the first transformer in a scikit-learn pipeline. It
    will receive a pandas DataFrame as input, and the names of the columns that you want to extract
    from the DataFrame for later use in the pipeline by other transformers.

    Parameters
    ----------
    column_names: list of str or str
        List containing the names of the columns that will be extracted
        from the input DataFrame. If you only want to extract one column,
        a single string can be passed instead.

    Examples
    --------
    >>> import pandas as pd
    >>> from sklearn.feature_extraction.text import CountVectorizer
    >>> from sklearn.pipeline import Pipeline
    >>> from pylat.wrapper.transformer.dataframe_selector import DataFrameSelector
    >>> df = pd.DataFrame(data={
    ...    'text': ['Sample text', 'another one', 'more text data', 'last one one'],
    ...    'labels': [0, 0, 1, 0],
    ...    'some_column': [0.01, 0.2, 0.1, 0.7]
    ... })
    >>> text_pipeline = Pipeline(steps=[
    ...     ('selector', DataFrameSelector('text')),
    ...     ('tf_idf', CountVectorizer())
    ... ])
    >>> print(text_pipeline.fit_transform(df).toarray())
    [[0 0 0 0 0 1 1]
     [1 0 0 0 1 0 0]
     [0 1 0 1 0 0 1]
     [0 0 1 0 2 0 0]]
    """

    def __init__(self, column_names):
        self.column_names = column_names

    def fit(self, x, y=None, **fit_params):
        return self

    def transform(self, x):
        """Extract the given column_names from x.

        :param x: pandas DataFrame from which the columns will be extracted.
        :return: numpy array containing the extracted columns.
        :raise InvalidArgumentError: If x is not a pandas DataFrame.
        :raise KeyError: If the column names passed in the constructor of the
            selector don't exist in the given DataFrame.
        """
        if not isinstance(x, pd.DataFrame):
            raise InvalidArgumentError('x', 'Transform must receive a pandas Dataframe as argument')

        return x[self.column_names].values



import logging
import numpy as np
import tensorflow as tf

from abc import ABC, abstractmethod
from sklearn.model_selection import train_test_split
from .utils import get_next_batch


class BaseTrainer(ABC):
    """

    Parameters
    ----------
    num_epochs: int (default=200)
        Maximum number of epochs taken during the training phase.

    batch_size: int (default=30)
        Number of training instances used for each gradient update
        in the training phase.
    """

    def __init__(self, model, num_epochs=200, batch_size=30):
        self.model = model
        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self.X_train = []
        self.y_train = []
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.saver = None

    def train(self, X, y):
        self.logger.info('Creating model...')
        self.model.init_model(X, y)
        self.saver = tf.train.Saver()

        self.logger.info('Starting training...')
        self._train_loop(X, y)
        self.on_train_finished()

    def _train_loop(self, X, y):
        self.on_train_loop_started(X, y)
        num_batches = len(self.X_train) // self.batch_size + 1
        for epoch in range(self.num_epochs):
            stop_train = self._epoch_loop(epoch, num_batches)
            if stop_train:
                break

    def _epoch_loop(self, num_epoch, num_batches):
        self.logger.info('Epoch: {}'.format(num_epoch))
        train_accuracy = 0
        train_loss = 0
        for batch_idx in range(num_batches):
            x_batch, y_batch = get_next_batch(self.X_train, self.y_train,
                                              batch_idx, self.batch_size)
            feed_dict = {
                self.model.x_t: x_batch,
                self.model.y_t: y_batch,
                self.model.is_training: True
            }
            _, batch_loss, batch_accuracy = self.model.session.run(
                [self.model.optimize, self.model.loss, self.model.error],
                feed_dict=feed_dict)
            train_accuracy += batch_accuracy
            train_loss += batch_loss
        train_accuracy /= num_batches
        train_loss /= num_batches
        return self.on_epoch_finished(train_loss, train_accuracy)

    def on_train_loop_started(self, X, y):
        self.X_train = X
        self.y_train = y

    def on_epoch_finished(self, loss, acc):
        self.logger.info('Train: Loss = {:.3f} - Accuracy = {:.3f}'.format(
            loss, acc * 100))
        return False

    def on_train_finished(self):
        pass


class EarlyStoppingTrainer(BaseTrainer):
    def __init__(self, model, batch_size, num_epochs,
                 validation_split=0.2, validation_data=None,
                 max_epochs_no_progress=5, random_seed=42):
        super().__init__(model, batch_size, num_epochs)
        self.validation_split = validation_split
        self.validation_data = validation_data
        self.max_epochs_without_progress = max_epochs_no_progress
        self.random_seed = random_seed
        self.best_loss = np.inf
        self.epochs_without_progress = 0
        self.X_val = []
        self.y_val = []

    def on_train_loop_started(self, X, y):
        if self.validation_data is None:
            self.X_train, self.X_val, self.y_train, self.y_val = \
                train_test_split(X, y, test_size=self.validation_split,
                                 random_state=self.random_seed)
        else:
            self.X_val = self.validation_data[0]
            self.y_val = self.validation_data[1]
            self.X_train = X
            self.y_train = y

    def on_epoch_finished(self, loss, acc):
        val_loss, val_acc = self.model.session.run(
            [self.model.loss, self.model.error],
            feed_dict={self.model.x_t: self.X_val, self.model.y_t: self.y_val})

        if val_loss < self.best_loss:
            self.best_loss = val_loss
            self.epochs_without_progress = 0
            self.saver.save(self.model.session, self.model.save_path)
        else:
            self.epochs_without_progress += 1
            if self.epochs_without_progress > self.max_epochs_without_progress:
                self.logger.info('No progress after %s epochs. '
                                 'Stopping...',
                                 self.max_epochs_without_progress)
                return True
        self.logger.info('Validation: Loss = {:.5f} - Best loss = '
                         '{:.5f} - Accuracy = {:.3f}'.format(val_loss,
                                                             self.best_loss,
                                                             val_acc * 100))
        return super().on_epoch_finished(loss, acc)

    def on_train_finished(self):
        self.saver.restore(self.model.session, self.model.save_path)



from sklearn.base import BaseEstimator, TransformerMixin

import abc


class BaseTransformer(abc.ABC, BaseEstimator, TransformerMixin):
    @abc.abstractmethod
    def fit(self, x, y=None):
        pass

    @abc.abstractmethod
    def transform(self, x):
        pass


class BasePredictor(abc.ABC, BaseEstimator):
    @abc.abstractmethod
    def predict(self, x, y):
        pass


class DataPipeline(BaseTransformer, BasePredictor):
    def __init__(self, transformers, predictor=None):
        self.transformers = transformers
        self.predictor = predictor
        self._assert_valid_params()

    def fit(self, x, y=None):
        for transformer in self.transformers:
            transformer.fit(x, y)

    def transform(self, x):
        transformed_x = x
        del x
        for transformer in self.transformers:
            transformed_x = transformer.transform(transformed_x)
        return transformed_x

    def predict(self, x, y):
        if self.predictor is None:
            raise Exception("The pipeline must have a predictor in order to call its predict method.")
        return self.predictor.predict(x, y)

    def _assert_valid_params(self):
        for transformer in self.transformers:
            assert(hasattr(transformer, "fit"))
            assert(hasattr(transformer, "transform"))

        assert(hasattr(self.predictor, "predict"))



from gensim.models import KeyedVectors
from .embeddings import Word2VecEmbedding
from pathlib import Path

import functools
import os
import tensorflow as tf

he_init = tf.contrib.layers.variance_scaling_initializer()


def get_next_batch(x, y, batch_idx, batch_size):
    """ Fetch next batch from data set for training.

    :param x: Feature matrix of size (num_samples, num_features).
    :param y: Label matrix of size (num_samples, 1)
    :param batch_idx: Index of the current batch.
    :param batch_size: Size of the batch
    :return: Tuple containing the current x and y batch.
    """
    start = batch_idx * batch_size
    end = (batch_idx + 1) * batch_size
    end = min(end, len(x) + 1)
    return x[start:end], y[start:end]


def last_relevant(output, length):
    """ Get relevant data from the output of a dynamic RNN layer

    :param output: Output tensor of the RNN layer.
    :param length: Length of the inputs of the layer.
    :return: Tensor containing the relevant output data for each data instance.
    """
    batch_size = tf.shape(output)[0]
    rows_idx = tf.range(0, batch_size)
    col_idx = tf.maximum(1, length) - 1
    final_idx = tf.stack([rows_idx, col_idx], axis=1)
    relevant = tf.gather_nd(output, final_idx)
    return relevant


def get_length_tensor(sequence):
    """ Return a tensor that computes the length of a sequence.

    This function can be used when working with padded textual
    data. It computes the actual length of each text input,
    without taken into account the padded elements.
    :param sequence: Tensor containing the input data.
    :return: Tensor that computes the actual length of the data.
    """
    used = tf.sign(tf.reduce_max(tf.abs(sequence), 2))
    length = tf.reduce_sum(used, 1)
    return tf.cast(length, tf.int32)


def lazy_property(func):
    attribute = '_cache_' + func.__name__

    @property
    @functools.wraps(func)
    def decorator(self):
        if not hasattr(self, attribute):
            setattr(self, attribute, func(self))
        return getattr(self, attribute)

    return decorator


def convert_vec_embedding(input_file, output_dir):
    """Converts a .vec file into the format used by the Embedding classes.

    Parameters
    ----------
    input_file
    output_dir

    Returns
    -------

    """
    path = Path(input_file)
    file_name = path.stem
    model = KeyedVectors.load_word2vec_format(input_file, unicode_errors='ignore')
    embedding = Word2VecEmbedding(model=model)
    embedding.save_embeddings(os.path.join(output_dir, file_name))



from pylat.wrapper.transformer.dataframe_selector import DataFrameSelector
from pylat.exceptions import InvalidArgumentError

import numpy as np
import pandas as pd
import pytest


class TestDataFrameSelector():
    def test_valid_df(self):
        df = pd.DataFrame(data={
            'important': [50, 2, 7],
            'not_important': [1, 3, 5]
        })
        selector = DataFrameSelector(column_names='important')
        res = selector.fit_transform(df)
        assert np.array_equal(res, [50, 2, 7])

    def test_invalid_df(self):
        df = 'Hi :)'
        selector = DataFrameSelector(column_names=['important'])
        with pytest.raises(InvalidArgumentError):
            selector.fit_transform(df)

    def test_valid_selection(self):
        df = pd.DataFrame(data={
            'a_column': [1, 2, 4, 1],
            'labels': ['Y', 'N', 'Y', 'N/A'],
            'text': ['Helloooo', 'This is a sample', 'Nice', 'Can you come here?'],
            'another_column': [0.45, 0.1213, 0.134, 0.0001]
        })
        selector = DataFrameSelector(column_names=['labels', 'text'])
        res = selector.fit_transform(df)
        assert np.array_equal(res, [
            ['Y', 'Helloooo'],
            ['N', 'This is a sample'],
            ['Y', 'Nice'],
            ['N/A', 'Can you come here?']
        ])

    def test_invalid_selection(self):
        df = pd.DataFrame(data={
            'a_column': [1, 2, 4, 1]
        })
        selector = DataFrameSelector(column_names=[2, 'muahahahah'])
        with pytest.raises(KeyError):
            selector.fit_transform(df)



from ..model import BaseNeuralNetwork
from tensorflow.python.saved_model import tag_constants

import numpy as np
import tensorflow as tf

import logging


class RecurrentNeuralNetwork(BaseNeuralNetwork):
    """Recurrent neural network implementation using TensorFlow.

    Parameters
    ----------
    num_epochs: int (default=100)
        Maximum number of epochs taken during the training phase.

    batch_size: int (default=200)
        Number of training instances used for each gradient update
        in the training phase.

    num_units: tuple or list of ints (default=(256,))
        Tuple containing the number of units that each recurrent layer will have. If
        the tuple contains more than one element, the recurrent layers will be stacked
        using a MultiRNNCell.

    cell_factory: BaseCellFactory object (default=GRUCellFactory())
        Instance of a cell factory that will be used to create the cells of each
        recurrent layer. For more info see :mod:`.rnn_cells`.

    activation: callable (default=None)
        Activation function applied to the output of each layer (e.g. tanh, ReLU...).

    kernel_initializer: callable (default=he_init)
        Function used to initialize the weights of the different layers.

    layer_norm: bool (default=False)
        Whether to apply layer normalization to each layer or not.

    dropout_rate: float (default=None)
        Dropout rate of each layer. Its value must be between 0 or 1.
        If its value is None or 0, no dropout is applied.

    learning_rate: float (default=1e-4)
        Learning rate used by the optimizer during the training phase.

    optimizer: tf.train.Optimizer class (default=tf.train.AdamOptimizer)
        Optimizer used during the training phase to minimize the loss
        function.

    Attributes
    ----------
    session : tf.Session
        TensorFlow session where the neural network operations will
        be executed.

    Examples
    --------
    >>> from pylat.neuralnet.rnn.recurrent_neural_network import \
        RecurrentNeuralNetwork
    >>> X = ['Hi, how are you doing?']
    >>> X_prep = X
    """

    def __init__(self,  rnn_layers, fc_layers, learning_rate=1e-4,
                 optimizer=tf.train.AdamOptimizer, save_dir='results',
                 random_seed=42, embeddings=None):
        super().__init__(save_dir)
        self.rnn_layers = rnn_layers
        self.fc_layers = fc_layers
        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.random_seed = random_seed
        self.w2v_embeddings = embeddings
        self.softmax = None
        self.loss_op = None
        self.train_op = None
        self.accuracy = None
        self._num_classes = 0
        self._num_steps = 0

    def init_model(self, X, y):
        self._check_input(X, y)
        self.logger.info('Building graph. Num steps: %s', self._num_steps)
        tf.set_random_seed(self.random_seed)
        super().init_model(X, y)

    @property
    def prediction(self):
        if self.softmax is None:
            with tf.name_scope('input'):
                rnn_inputs = self._input()

            with tf.name_scope('dnn'):
                logits = self._rnn(rnn_inputs)
                softmax = tf.nn.softmax(logits, name='y_proba')

            self._logits = logits
            self.softmax = softmax
        return self.softmax

    @property
    def loss(self):
        if self.loss_op is None:
            with tf.name_scope('loss'):
                xent = tf.keras.losses.sparse_categorical_crossentropy(self.y_t,
                                                                self.softmax)
                self.loss_op = tf.reduce_mean(xent, name='loss')
        return self.loss_op

    @property
    def optimize(self):
        if self.train_op is None:
            with tf.name_scope('train'):
                optimizer = self.optimizer(self.learning_rate)
                self.train_op = optimizer.minimize(self.loss)
        return self.train_op

    @property
    def error(self):
        if self.accuracy is None:
            with tf.name_scope('accuracy'):
                y_pred = tf.argmax(self.prediction, axis=1, name='output')
                correct = tf.equal(y_pred, self.y_t)
                self.accuracy = tf.reduce_mean(tf.cast(correct, dtype=tf.float32))
        return self.accuracy

    def _check_input(self, X, y):
        self._num_classes = len(np.unique(y))
        self._num_steps = len(max(X, key=len))

    def _rnn(self, inputs):
        current_input = inputs
        self.is_training = tf.placeholder_with_default(False, shape=(),
                                                       name='is_training')
        kwargs = {'training': self.is_training}
        with tf.name_scope('rnn_layers'):
            for layer in self.rnn_layers:
                current_input = layer.build_tensor(current_input, **kwargs)
        with tf.name_scope('fc_layers'):
            for layer in self.fc_layers:
                current_input = layer.build_tensor(current_input, **kwargs)
            output = tf.keras.layers.Dense(self._num_classes)(current_input)
        return output

    def _input(self):
        self.x_t = tf.placeholder(tf.int64, shape=[None, self._num_steps],
                                  name='x_input')
        self.y_t = tf.placeholder(tf.int64, shape=[None], name='y_input')
        with tf.name_scope('embeddings'):
            np_emb = np.array(self.w2v_embeddings.get_vectors())
            vocab_size, emb_dim = np.shape(np_emb)
            embeddings = tf.Variable(
                tf.constant(0.0, shape=[vocab_size, emb_dim]),
                trainable=False, name="W")
            self._emb_placeholder = tf.placeholder(tf.float32,
                                                   [vocab_size, emb_dim],
                                                   name='emb_plh')
            self._emb_init = embeddings.assign(self._emb_placeholder,
                                               name='emb_init')
            rnn_inputs = tf.nn.embedding_lookup(embeddings, self.x_t)
        return rnn_inputs

    def additional_inits(self, **params):
        self.session.run(self._emb_init, feed_dict={
            self._emb_placeholder: self.w2v_embeddings.get_vectors(**params)
        })

    def save(self, save_path):
        inputs = {"x_t": self.x_t}
        outputs = {"pred_proba": self.prediction}
        tf.saved_model.simple_save(self.session, save_path, inputs, outputs)

    def restore(self, save_path, **params):
        graph = tf.Graph()
        self.session = tf.Session(graph=graph)
        tf.saved_model.loader.load(
            self.session,
            [tag_constants.SERVING],
            save_path,
        )
        self.x_t = graph.get_tensor_by_name('input/x_input:0')
        self.softmax = graph.get_tensor_by_name('dnn/y_proba:0')
        self._emb_init = graph.get_tensor_by_name('input/embeddings/emb_init:0')
        self._emb_placeholder = graph.get_tensor_by_name('input/embeddings/'
                                                         'emb_plh:0')



from pylat.wrapper.transformer.base import BaseTransformer
from sklearn.base import TransformerMixin

import numpy as np


class WordEmbeddingsTransformer(BaseTransformer, TransformerMixin):
    """Transform a list of tokens to numbers using a word embedding.

    This is a wrapper around word2vec, complying to the scikit-learn
    transformer API, which can be used in a transformer pipeline to
    transform some input textual data to word vectors.

    The word vectors are learned based on the parameters given
    in the constructor of the embedding, allowing its use for hyper-parameter
    tuning (e.g. RandomizedSearchCV).

    Parameters
    ----------
    embeddings : :obj:`BaseWordEmbedding`
        Instance of BaseWordEmbedding class that will be used to train the
        embeddings and transform the tokens.
    fit_corpus : :obj:`list` of :obj:`list` of str, optional (default=None)
        A 2 dimensional list, where the first dimension contains every sentence
        used to train the embedding, and the second dimension contains every
        token of each sentence. If set to None, the sentences to be transformed
        will be fed as training data to the embedding.
    to_id : bool, optional (default=True)
        If set to True, each token will be transformed to their id in the
        embedding. If set to False, each token will be transformed to their
        vector representation instead.

    Examples
    --------
    >>> from pylat.wrapper.transformer.embeddings_transformer import \
            WordEmbeddingsTransformer
    >>> from pylat.neuralnet.embeddings import GensimConfig, Word2VecEmbedding
    >>> X = [['Sample', 'text'], ['another', 'one'], ['last', 'one']]
    >>> # arguments passed in order to make execution deterministic
    >>> config = GensimConfig(size=2, epochs=5, workers=1, seed=42, \
                              min_count=1, window=1, hashfxn=len)
    >>> embedding = Word2VecEmbedding(config)
    >>> vec_transformer = WordEmbeddingsTransformer(embedding, to_id=False)
    >>> vec_transformer.fit(X).transform(X)
        array([[[ 0.18671471,  0.23427033], [ 0.19643007, -0.08401009]],
               [[-0.24481292,  0.0009373 ], [-0.13900341,  0.18536615]],
               [[ 0.19643007, -0.08401009], [-0.13900341,  0.18536615]]],
               dtype=float32)
    >>> id_transformer = WordEmbeddingsTransformer(embedding)
    >>> id_transformer.fit(X).transform(X)
        array([[1, 2],
               [3, 0],
               [4, 0]])
    """

    def __init__(self, embeddings, fit_corpus=None, to_id=True):
        self.embeddings = embeddings
        self._fit_corpus = fit_corpus
        self._to_id = to_id

    def fit(self, x, y=None, **fit_params):
        """Fits the word embeddings to the given data

        Note
        ----
        If a fit_corpus was passed in the constructor of this class, that data
        will be used to fit the embeddings instead.

        Parameters
        ----------
        x : :obj:`list` of :obj:`list` of str
            A 2 dimensional list, where the first dimension contains every
            sentence used to train the embedding, and the second dimension
            contains every token of each sentence. If set to None, the sentences
            to be transformed will be fed as training data to the embedding.
        y : :obj:`list`, optional (default=None)
            Labels of the passed data.
        fit_params :

        Returns
        -------
        self
            Reference to the class after being trained.
        """
        train_data = x if self._fit_corpus is None else self._fit_corpus
        self.embeddings.train(train_data, **fit_params)
        return self

    def transform(self, x, **transf_params):
        """Transforms the input text into numbers.

        This method uses the embeddings trained in the fit method to

        Parameters
        -----------
        x : :obj:`list` of :obj:`list` of str
            2 dimensional list, containing the tokens of every sentence.

        Returns
        -------
        return : numpy array of shape of shape [n, ]
            Each row in the output array corresponds to each document in the
            input array, and it contains a list of preprocessed tokens.
        """
        func = self.embeddings.to_id if self._to_id \
            else self.embeddings.to_vector
        return np.asarray([[func(token, **transf_params) for token in sentence
                           if func(token, **transf_params) is not None]
                           for sentence in x])

    def fit_transform(self, X, y=None, **params):
        return self.fit(X, **params).transform(X, **params)


class DocumentEmbeddingsTransformer(BaseTransformer):
    """Transforms a list of sentences into their corresponding embeddings.

    This is a wrapper around doc2vec, complying to the scikit-learn
    transformer API, which can be used in a transformer pipeline to
    transform some input textual data to document vectors.

    The document vectors are learned based on the parameters given
    in the constructor, allowing its use for hyper-parameter tuning
    (e.g. RandomizedSearchCV).

    Parameters
    ----------
    embeddings : :obj:`Doc2VecEmbedding`
        Instance of the Doc2VecEmbedding class used to transform the sentences.
    fit_corpus : :obj:`list` of :obj:`list` of str, optional (default=None)
        A 2 dimensional list, containing the tokens of every sentence that will
        be used to train the embeddings. If set to None, the data to be
        transformed into vectors will be also used as training data.

    Examples
    --------
    >>> from pylat.wrapper.transformer.embeddings_transformer import \
            DocumentEmbeddingsTransformer
    >>> from pylat.neuralnet.embeddings import GensimConfig, Doc2VecEmbedding
    >>> X = [['Sample', 'text'], ['another', 'one'], ['last', 'one']]
    >>> # arguments passed in order to make execution deterministic
    >>> config = GensimConfig(size=2, epochs=5, workers=1, seed=42, \
                              min_count=1, window=1, hashfxn=len)
    >>> embedding = Doc2VecEmbedding(config)
    >>> d2v = DocumentEmbeddingsTransformer(embedding, fit_corpus=X)
    >>> d2v.fit(X).transform(X)
    array([[-0.15986516, -0.24026237],
           [-0.15986516, -0.24026237],
           [ 0.18679854,  0.23427881]], dtype=float32)
    """

    def __init__(self, embeddings, fit_corpus=None):
        self.embeddings = embeddings
        self._fit_corpus = fit_corpus

    def fit(self, x, y=None):
        """

        Parameters
        ----------
        x
        y

        Returns
        -------

        """
        train_data = x if self._fit_corpus is None else self._fit_corpus
        self.embeddings.train(train_data)
        return self

    def transform(self, x):
        """

        Parameters
        ----------
        x :

        Returns
        -------

        """
        return np.asarray([self.embeddings.to_vector(sentence)
                           for sentence in x])



import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from pylat.exceptions import InvalidArgumentError

__author__ = 'Alejandro González Hevia'


class SentencePadder(BaseEstimator, TransformerMixin):
    """Pads sentences of word vectors for use in a recurrent neural network.

    After preprocessing textual data and creating a vector representation of each word,
    each input will have different dimensions (one per word). In order to use the data
    in a recurrent neural network we need to pad every training instance with zeros up
    to a certain dimension. This dimension can be the greatest of the training instances
    or another one.

    Parameters
    ----------
    padding_length: int (default=None)
        Output length of each sample instance. If set to None, the output length
        will be equal to the max length from the sample set.

    Examples
    --------
    >>> from pylat.wrapper.transformer.sentence_padder import SentencePadder
    >>> X = [
    ...         [1, 4, 5, 2, 3, 5, 5, 1, 3],
    ...         [1, 2, 3],
    ...         [4, 5, 1, 2, 3, 1]
    ...     ]
    >>> default_padder = SentencePadder()
    >>> default_padder.fit_transform(X)
        array([[1., 4., 5., 2., 3., 5., 5., 1., 3.],
               [1., 2., 3., 0., 0., 0., 0., 0., 0.],
               [4., 5., 1., 2., 3., 1., 0., 0., 0.]], dtype=float32)
    >>> custom_padder = SentencePadder(padding_length=15)
    >>> custom_padder.fit_transform(X)
        array([[1., 4., 5., 2., 3., 5., 5., 1., 3., 0., 0., 0., 0., 0., 0.],
               [1., 2., 3., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
               [4., 5., 1., 2., 3., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0.]], dtype=float32)
    """

    def __init__(self, padding_length=None):
        self.padding_length = padding_length

    def fit(self, x, y=None, **fit_params):
        max_length = len(max(x, key=len))
        if self.padding_length is not None and self.padding_length < max_length:
            raise InvalidArgumentError('padding_length', 'Padding length must be greater \
                                                          or equal to the maximum sentence length.')
        elif self.padding_length is None:
            self.padding_length = max_length
        return self

    def transform(self, x):
        num_instances = len(x)
        ret = np.zeros(shape=[num_instances, self.padding_length],
                       dtype=np.float32)
        for idx, sentence in enumerate(x):
            if len(sentence) != 0:
                ret[idx, :len(sentence)] = sentence
        return ret



from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError
from pylat.exceptions import InvalidArgumentError

import en_core_web_sm, es_core_news_sm

import numpy as np

import copy
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

id2pkg = {
    'en': en_core_web_sm,
    'es': es_core_news_sm
}


class TextPreprocessor(BaseEstimator, TransformerMixin):
    """Basic text preprocessing class.

    This class provides several options for the preprocessing and
    cleaning of textual data, and it serves as a good first step before
    training models that work with this kind of data.

    It implements the BaseEstimator and TransformerMixin classes from
    sklearn, so it can be used in its pipelines and all of its hyperparameter
    search classes.

    Parameters
    ----------
    spacy_model_id : str, optional (default='en')
        Id of the Spacy model that will be used to perform the basic
        natural language operations (parser, tagger, ner...). The model
        must have been imported first (see https://spacy.io/usage/models).

    additional_pipes : iterable or None, optional (default=None)
        Iterable of callables that will be appended to the original pipeline
        of the Spacy model. This callables receive a Doc object as input,
        and return another Doc object.

    remove_stop_words : boolean, optional (default=False)
        Whether to remove stop_words from the input text or not.

    lemmatize : boolean, optional (default=False)
        Whether to return a lemmatized form of each token from the input text or not.

    Attributes
    ----------
    nlp : spacy.Language
        Loaded language model matching the received model id.

    custom_nlp : spacy.Language
        Final language model used for preprocessing, which includes
        the user defined pipes added in the 'additional_pipes' param.

    model_caches : dict,
        Dictionary that keeps a cache of all the loaded nlp models. Its keys
        are the string model ids, and its values are their respective nlp models.

    Examples
    --------
    >>> from pylat.wrapper.transformer.text_preprocessor import TextPreprocessor
    >>> X = ['Hi, how are you doing?']
    >>> base_preprocessor = TextPreprocessor()
    >>> print(base_preprocessor.fit_transform(X))
    [['hi' ',' 'how' 'are' 'you' 'doing' '?']]
    >>> lemmatizer = TextPreprocessor(lemmatize=True)
    >>> print(lemmatizer.fit_transform(X))
    [['hi' ',' 'how' 'be' 'you' 'do' '?']]
    >>> preprocessor = TextPreprocessor(remove_stop_words=True)
    >>> print(preprocessor.fit_transform(X))
    [['hi']]
    """

    model_caches = {}

    def __init__(self, spacy_model_id='en', additional_pipes=None,
                 remove_stop_words=False, lemmatize=False,
                 disable=(), custom_nlp=None):
        self.spacy_model_id = spacy_model_id
        self.remove_stop_words = remove_stop_words
        self.lemmatize = lemmatize

        if additional_pipes is None:
            additional_pipes = []
        self.additional_pipes = additional_pipes
        self.disable = disable
        self.custom_nlp = custom_nlp

    @property
    def nlp(self):
        """Spacy language model loaded given the id passed in the constructor.

        A dict of cached models is updated in order to avoid the expensive operation
        of loading a Language object.

        :return: spacy.Language object
            Spacy language model.
        """
        if self.spacy_model_id not in self.model_caches:
            logger.info('Model {} was not found in cache'.format(self.spacy_model_id))
            logger.info('Loading model {}...'.format(self.spacy_model_id))
            pkg = id2pkg[self.spacy_model_id]
            self.model_caches[self.spacy_model_id] = pkg.load(disable=self.disable)

        return self.model_caches[self.spacy_model_id]

    def fit(self, x, y=None):
        """Fit to data.

        This method must be called before transform. It doesn't take
        into account any of the data received as parameters, but it
        validates the params passed in the constructor, following
        sklearn conventions.

        :param x: any
            Not taken into account.
        :param y: any (default=None)
            Not taken into account.
        :return: self
        """
        self._assert_valid_types()

        # add user defined pipes to base nlp object
        self.custom_nlp = copy.deepcopy(self.nlp)
        for pipe in self.additional_pipes:
            logger.debug('Adding pipe "{}" to nlp pipeline.'.format(pipe))
            self.custom_nlp.add_pipe(pipe)

        return self

    def transform(self, x):
        """Transforms the input text into preprocessed data.

        :param x: array of strings of any shape (n)
        :return: numpy array of shape of shape [n, ]
            Each row in the output array corresponds to each document in the
            input array, and it contains a list of preprocessed tokens.
        """
        if self.custom_nlp is None:
            raise NotFittedError

        ret = []
        for sentence in x:
            doc = self.custom_nlp(sentence)
            ret.append(self._process(doc))

        return np.asarray(ret)

    def _process(self, doc):
        tokens = np.asarray(doc)
        if self.remove_stop_words:
            tokens = [tok for tok in tokens if not tok.is_stop and not tok.is_punct]

        if self.lemmatize:
            tokens = [tok.lemma_.lower().strip() if tok.lemma_ != '-PRON-'
                      else tok.text.lower().strip() for tok in tokens]
        else:
            tokens = [tok.text.lower().strip() for tok in tokens]
        return tokens

    def get_params(self, deep=True):
        return {
            "spacy_model_id": self.spacy_model_id,
            "custom_nlp": self.custom_nlp,
            "remove_stop_words": self.remove_stop_words,
            "lemmatize": self.lemmatize,
            "additional_pipes": self.additional_pipes,
            "disable": self.disable
        }

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            if parameter == 'additional_pipes' and value is None:
                setattr(self, parameter, [])
                continue

            setattr(self, parameter, value)
        return self

    def _assert_valid_types(self):
        if not isinstance(self.remove_stop_words, bool):
            raise InvalidArgumentError('remove_stop_words', 'Remove stop words option must be True or False.')

        if not isinstance(self.lemmatize, bool):
            raise InvalidArgumentError('lemmatize', 'Lemmatize option must be True or False.')

        if not hasattr(self.additional_pipes, '__iter__'):
            logger.info('Additional pipes is: {}'.format(self.additional_pipes))
            raise InvalidArgumentError('additional_pipes', 'Additional pipes must be an iterable of callables')

        if self.spacy_model_id not in id2pkg.keys():
            raise InvalidArgumentError('spacy_model_id',
                                       'Model not found. List of valid model '
                                       'ids: {}'.format(id2pkg.keys()))

        for pipe in self.additional_pipes:
            if not hasattr(pipe, '__call__'):
                raise InvalidArgumentError('additional_pipes', "All additional pipes must be callables")



import logging
import numpy as np
import tensorflow as tf

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.exceptions import NotFittedError

from pylat.neuralnet.rnn.recurrent_neural_network import RecurrentNeuralNetwork
from pylat.neuralnet.trainer import BaseTrainer, EarlyStoppingTrainer


class RNNWrapper(BaseEstimator, ClassifierMixin):

    def __init__(self, rnn_layers, fc_layers, num_epochs=30,
                 batch_size=200, learning_rate=1e-4,
                 optimizer=tf.train.AdamOptimizer, save_dir='results',
                 random_seed=42, embeddings=None, early_stopping=False,
                 validation_split=0.2, validation_data=None,
                 max_epochs_no_progress=5):
        self.model = RecurrentNeuralNetwork(rnn_layers, fc_layers,
                                            learning_rate, optimizer,
                                            save_dir, random_seed, embeddings)
        self.trainer = BaseTrainer(self.model, num_epochs, batch_size) \
            if not early_stopping \
            else EarlyStoppingTrainer(self.model, num_epochs, batch_size,
                                      validation_split, validation_data,
                                      max_epochs_no_progress)
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)

    def fit(self, X, y, **fit_args):
        self.trainer.train(X, y)

    def predict(self, x, **params):
        self.logger.info('Predict x: %s', np.shape(x))
        if self.model.session is None:
            raise NotFittedError
        else:
            probabilities = self._predict_proba(x, **params)
            return np.argmax(probabilities, axis=1)

    def _predict_proba(self, x, **params):
        self.logger.info('Predict probabilities of x: %s', np.shape(x))
        with self.model.session.as_default():
            self.model.additional_inits(**params)
            return self.model.prediction.eval(feed_dict={
                self.model.x_t: x
            })



import tensorflow as tf

from ..layers import BaseLayer
from .cells import GRUCellFactory


class RecurrentLayer(BaseLayer):
    def __init__(self, num_units, kernel_init='glorot_uniform',
                 activation='relu', dropout_rate=None,
                 cell_factory=GRUCellFactory(), layer_norm=False,
                 cell_dropout=0.0):
        super().__init__(num_units, kernel_init, activation, dropout_rate)
        self.cell_factory = cell_factory
        self.layer_norm = layer_norm
        self.cell_dropout = cell_dropout

    def build_tensor(self, inputs, **kwargs):
        final_cell = self.cell_factory(self.num_units, self.activation,
                                       self.kernel_init, self.cell_dropout,
                                       self.layer_norm)
        rnn_layer = tf.keras.layers.RNN(final_cell)(inputs, **kwargs)
        if self.dropout_rate is not None:
            rnn_layer = tf.keras.layers.Dropout(self.dropout_rate)(rnn_layer,
                                                                   **kwargs)
        return rnn_layer


class BidirectionalRecurrentLayer(RecurrentLayer):
    def build_tensor(self, inputs, **kwargs):
        final_cell = self.cell_factory(self.num_units, self.activation,
                                       self.kernel_init, self.cell_dropout,
                                       self.layer_norm)
        rnn = tf.keras.layers.RNN(final_cell)
        bidi = tf.keras.layers.Bidirectional(rnn)(inputs, **kwargs)
        if self.dropout_rate is not None:
            bidi = tf.keras.layers.Dropout(self.dropout_rate)(bidi, **kwargs)
        return bidi



from abc import ABC, abstractmethod
from pylat.exceptions import InvalidArgumentError
from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from gensim.models.word2vec import Word2Vec
from sklearn.exceptions import NotFittedError

import codecs
import logging
import numpy as np
import os

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class BaseEmbedding(ABC):
    """Base class of all the embeddings.

    This class provides the two methods that all the embedding classes provide
    to the user. It can be extended by other classes to add extra functionality.
    """

    @abstractmethod
    def train(self, train_data):
        """Train the embedding with the given training data.

        Parameters
        ----------
        train_data

        Returns
        -------
        self
            Trained embedding instance.
        """
        pass

    @abstractmethod
    def to_vector(self, text):
        """Transforms text into its numerical representation

        Parameters
        ----------
        text

        Returns
        -------
        list of float
            Vector representation of the input text.
        """
        pass


class GensimConfig:
    """ Configuration class for embeddings from the gensim module.

    This class stores the parameters needed to train embeddings from the
    gensim module ('word2vec' and 'doc2vec'). It is used by the
    Word2VecEmbedding and Doc2VecEmbedding classes in order to train their
    respective models.

    Parameters
    ----------
    epochs : int, optional (default=100)
        Number of epochs used to learn the vector representations.

    size : int, optional (default=100)
        Number of dimensions that the final vector representation will have.

    min_count : int, optional (default=5)
        Minimum number of occurrences of a word in the training data in
        order to be added to the learned vocabulary.

    max_vocab_size : int, optional (default=None)
        Maximum size of the vocabulary learned. Most infrequent words will
        not be added to the vocab in order to fulfill this constraint. If None,
        all the words that have at least min_count occurrences will be added
        to the vocabulary.

    window : int, optional (default=5)
        Number of neighbouring words used to learn the vector representations.

    alpha : float, optional (default=0.025)
        Initial learning rate of the model.

    min_alpha : float, optional (default=0.0001)
        Minimum learning rate that the model will reach during training.

    sample : float, optional (default=0.01)
        This value configures which words with higher frequency are downsampled.

    seed : int, optional (default=42)
        Random seed used internally by the model in the training phase.

    workers : int, optional (default=3)
        Number of threads used to train the model.

    hashfxn : func, optional (default=hash)
        Function used internally to perform hashing of the words. If you want
        to obtain completely deterministic results, you can set this function
        to `len`. Otherwise, it is recommended to leave the default value.
    """
    def __init__(self, epochs=100, size=100, min_count=5, max_vocab_size=None,
                 window=5, alpha=0.025, min_alpha=0.0001, sample=0.001,
                 seed=42, workers=3, hashfxn=hash):
        self.size = size
        self.alpha = alpha
        self.window = window
        self.min_alpha = min_alpha
        self.min_count = min_count
        self.max_vocab_size = max_vocab_size
        self.sample = sample
        self.seed = seed
        self.workers = workers
        self.iter = epochs
        self.hashfxn = hashfxn

    def to_doc2vec(self):
        """Transforms the configuration to the specific Doc2Vec class names.

        Returns
        -------
        dict
            Dictionary containing every parameter of the configuration used
            by the Doc2Vec Gensim class.
        """
        params_dict = vars(self)
        # change the following two params to be the same as the Doc2Vec ones
        params_dict['vector_size'] = params_dict.pop('size')
        params_dict['epochs'] = params_dict.pop('iter')
        return params_dict

    def to_word2vec(self):
        """Transforms the configuration to the specific Word2Vec class names.

        Returns
        -------
        dict
            Dictionary containing every parameter of the configuration used
            by the Word2Vec Gensim class.
        """
        return vars(self)


class Doc2VecEmbedding(BaseEmbedding):
    """Embedding class that transforms sentences into vectors.

    This class is a wrapper around the Gensim Doc2Vec implementation that
    implements the BaseEmbedding interface.

    Parameters
    ----------
    gensim_conf : :obj:`GensimConfig` (default=GensimConfig())
        An instance of the GensimConfig class providing the parameters that will
        be used to train the embeddings.

    model : Doc2Vec model (default=None)
        Doc2Vec loaded model to use. If None, a new model will be trained
        with the given GensimConfig. If you pass a model, the training step
        will be skipped, so the Gensim configuration will not be used.
    """

    def __init__(self, gensim_conf=GensimConfig(), model=None):
        self.model = model
        self.conf = gensim_conf

    def train(self, train_data):
        if self.model is None:
            _assert_valid_train_data(train_data)
            _assert_valid_conf(self.conf)
            tagged_data = [TaggedDocument(words=sent, tags=[idx]) for idx, sent
                           in enumerate(train_data)]
            self.model = Doc2Vec(documents=tagged_data,
                                 **self.conf.to_doc2vec())

    def to_vector(self, text):
        if self.model is None:
            raise NotFittedError
        return self.model.infer_vector(text)


class BaseWordEmbedding(BaseEmbedding, ABC):
    """Base class of all the word embedding classes.

    This class provides the basic interface of all the embeddings that work at
    a word level.
    """

    @abstractmethod
    def to_id(self, token):
        """Return the id of the given token in the embedding.

        Parameters
        ----------
        token : str
            Token for which the id will be retrieved.

        Returns
        -------
        int
            ID of the token if it is present in the embedding, None otherwise.
        """
        pass

    @abstractmethod
    def to_word(self, token_id):
        """Returns the word corresponding to the given id in the embedding.

        Parameters
        ----------
        token_id : int
            ID of the token that will be retrieved.

        Returns
        -------
        str
            Token with the given id if it is present in the embedding,
            None otherwise.
        """
        pass

    @abstractmethod
    def get_vectors(self):
        """Returns all the vectors computed by the embedding instance.

        Returns
        -------
        2d array of float
            Values of the numeric representations of every word.
        """
        pass


class Word2VecEmbedding(BaseWordEmbedding):
    def __init__(self, gensim_conf=GensimConfig(), model=None):
        self.conf = gensim_conf
        self.model = model

    def train(self, train_data):
        if self.model is None:
            _assert_valid_train_data(train_data)
            _assert_valid_conf(self.conf)
            self.model = Word2Vec(sentences=train_data,
                                  **self.conf.to_word2vec())

    def to_vector(self, token):
        try:
            return self.model.wv[token]
        except KeyError:
            return None

    def to_id(self, token):
        try:
            return self.model.wv.vocab[token].index
        except KeyError:
            return None

    def to_word(self, token_id):
        return self.model.wv.index2word[token_id]

    def get_vectors(self):
        return self.model.wv.vectors

    def save_embeddings(self, save_path):
        np.save(save_path + '.npy', self.model.wv.vectors)
        with codecs.open(save_path + '.vocab', 'w', 'utf-8') as f_out:
            for word in self.model.wv.index2word:
                f_out.write(word + '\n')


class CrossLingualPretrainedEmbedding(BaseWordEmbedding):
    def __init__(self, embeddings_dir):
        self.embeddings_dir = embeddings_dir
        self.embeddings_dict = {}

    def train(self, train_data, **params):
        pass

    def to_vector(self, token, **params):
        embeddings = self._get_embeddings_from(**params)
        try:
            index = embeddings.word2index[token]
            return embeddings.wv[index]
        except KeyError:
            return None

    def to_id(self, token, **params):
        try:
            return self._get_embeddings_from(**params).word2index[token]
        except KeyError:
            return None

    def to_word(self, token_id, **params):
        return self._get_embeddings_from(**params).index2word[token_id]

    def get_vectors(self, **params):
        embeddings = self._get_embeddings_from(**params)
        return embeddings.wv

    def _load_vector(self, language):
        vector_path = os.path.join(self.embeddings_dir, language)
        self._assert_files_exist(vector_path, language)
        with codecs.open(vector_path + '.vocab', 'r', 'utf-8') as f_in:
            index2word = [line.strip() for line in f_in]
        word2index = {w: i for i, w in enumerate(index2word)}
        wv = np.load(vector_path + '.npy')
        self.embeddings_dict[language] = self.LoadedEmbedding(wv, index2word,
                                                              word2index)

    def _get_embeddings_from(self, **params):
        if params.get('language') is None:
            logger.warning('Language parameter was not specified. Input '
                           'sentences are assumed to be in English.')
            language = 'en'
        else:
            language = params['language']

        if self.embeddings_dict.get(language) is None:
            logger.info("Embeddings not available in cache, loading them...")
            self._load_vector(language)

        return self.embeddings_dict[language]

    def _assert_files_exist(self, vector_path, language):
        if not os.path.exists(vector_path + '.vocab'):
            raise InvalidArgumentError(self.embeddings_dir, 'Could not find '
                                       'vocab file for language {} in given '
                                       'directory.'.format(language))
        if not os.path.exists(vector_path + '.npy'):
            raise InvalidArgumentError(self.embeddings_dir, 'Could not find '
                                       'npy file for language {} in given '
                                       'directory.'.format(language))

    class LoadedEmbedding:
        """Inner class that stores information about the preloaded embeddings.

        Parameters
        ----------

        """
        def __init__(self, wv, index2word, word2index):
            self.wv = wv
            self.index2word = index2word
            self.word2index = word2index


def _assert_valid_train_data(train_data):
    if not hasattr(train_data, '__iter__') or type(train_data) == str:
        raise InvalidArgumentError('train_data', 'Train data must be an '
                                                 'iterable of tokens.')
    elif len(train_data) == 0:
        raise InvalidArgumentError('train_data', "Train data can't be empty")


def _assert_valid_conf(gensim_conf):
    if not isinstance(gensim_conf, GensimConfig):
        raise InvalidArgumentError(gensim_conf, 'Configuration parameter '
                                   'must be an instance of the GensimConfig '
                                   'class.')
    elif gensim_conf.iter <= 0:
        raise InvalidArgumentError('epochs', 'Number of epochs must be greater '
                                             'than 0.')
    elif gensim_conf.window <= 0:
        raise InvalidArgumentError('window', 'Window size must be greater '
                                             'than 0.')
    elif gensim_conf.max_vocab_size is not None \
            and gensim_conf.max_vocab_size <= 0:
        raise InvalidArgumentError('max_vocab_size',
                                   'Maximum vocabulary size must be greater '
                                   'than 0.')
    elif gensim_conf.size <= 0:
        raise InvalidArgumentError('size', 'Vector size must be greater '
                                           'than 0.')



from pylat.neuralnet.embeddings import GensimConfig, Word2VecEmbedding, \
    Doc2VecEmbedding, CrossLingualPretrainedEmbedding
from pylat.exceptions import InvalidArgumentError
from gensim.models import Doc2Vec, Word2Vec

import numpy as np
import os
import pytest
import unittest


class TestEmbeddings(unittest.TestCase):
    def setUp(self):
        self.valid_input = [
            ["This", "is", "the", "first", "sentence"],
            ["And", "this", "is", "the", "second", "one"],
            ["Each", "sentence", "is", "a", "list", "of", "tokens"]
        ]
        self.invalid_input = "This is some text"
        self.embeddings_dir = os.path.join('test', 'data', 'embeddings')

    def test_pretrained_word2vec(self):
        model = Word2Vec.load(os.path.join(self.embeddings_dir,
                                           'w2v_test.model'))
        embeddings = Word2VecEmbedding(model=model)
        embeddings.train(None)  # shouldn't do anything in train
        self.assertEqual(embeddings.model, model)
        self.assertEqual(embeddings.to_id('pron'), 0)
        self.assertEqual(embeddings.to_id('invented'), None)

    def test_invalid_conf(self):
        invalid_configs = [GensimConfig(epochs=0), GensimConfig(window=0),
                           GensimConfig(size=0), GensimConfig(max_vocab_size=0)]
        for config in invalid_configs:
            embeddings = Word2VecEmbedding(gensim_conf=config)
            with pytest.raises(InvalidArgumentError):
                embeddings.train(self.valid_input)

    def test_to_id(self):
        config = GensimConfig(epochs=5, max_vocab_size=None, min_count=1)
        embeddings = Word2VecEmbedding(gensim_conf=config)
        embeddings.train(self.valid_input)
        self.assertEqual(embeddings.to_id('sentence'), 2)
        self.assertEqual(embeddings.to_word(2), 'sentence')

    def test_to_vector(self):
        config = GensimConfig(epochs=5, max_vocab_size=None,
                              min_count=1, size=5)
        w2v = Word2VecEmbedding(gensim_conf=config)
        w2v.train(self.valid_input)
        self.assertEqual(len(w2v.to_vector('sentence')), 5)
        d2v = Doc2VecEmbedding(gensim_conf=config)
        d2v.train(self.valid_input)
        self.assertEqual(len(d2v.to_vector(['my', 'sentence'])), 5)

    def test_pretrained_doc2vec(self):
        model = Doc2Vec.load(os.path.join(self.embeddings_dir,
                                          'd2v_test.model'))
        embeddings = Doc2VecEmbedding(model=model)
        embeddings.train(None)  # shouldn't do anything in train
        self.assertEqual(embeddings.model, model)

    def test_spanish_multilingual_embedding(self):
        model = CrossLingualPretrainedEmbedding(self.embeddings_dir)
        self.assertEqual(model.to_id('amigo', language='es'), 0)
        np.testing.assert_allclose(model.to_vector('amor', language='es'),
                                   np.asarray([0.7, 0.2, 0.4]))

    def test_change_language_multilingual_embedding(self):
        model = CrossLingualPretrainedEmbedding(self.embeddings_dir)
        self.assertEqual(model.to_id('tierno', language='es'), 4)
        self.assertEqual(model.to_id('tierno', language='en'), None)
        np.testing.assert_allclose(model.to_vector('cute', language='en'),
                                   np.asarray([0.3, 0.8, 0.6]))

    def test_no_language_set_multilingual_embedding(self):
        model = CrossLingualPretrainedEmbedding(self.embeddings_dir)
        self.assertEqual(model.to_id('amigo'), None)
        self.assertEqual(model.to_id('friend'), 2)

    def test_invalid_language_multilingual_embedding(self):
        model = CrossLingualPretrainedEmbedding(self.embeddings_dir)
        with pytest.raises(InvalidArgumentError):
            model.to_id('친구', language='ko')

    def test_save_embeddings(self):
        conf = GensimConfig(size=2, min_count=1)
        embeddings = Word2VecEmbedding(conf)
        embeddings.train(self.valid_input)
        embeddings.save_embeddings('emb')
        emb_arr = np.load('emb.npy')
        self.assertEqual(np.shape(emb_arr), (14, 2))
        with open('emb.vocab', 'r') as f:
            vocab = f.readlines()
            self.assertEqual(len(vocab), 14)
        os.remove('emb.npy')
        os.remove('emb.vocab')

    def test_invalid_input(self):
        conf = GensimConfig(min_count=1)
        embeddings = Word2VecEmbedding(conf)
        with pytest.raises(InvalidArgumentError):
            embeddings.train(self.invalid_input)


if __name__ == '__main__':
    unittest.main()



from sklearn.base import BaseEstimator, TransformerMixin


class ToArrayTransformer(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X.toarray()



from setuptools import setup

setup(
    name='pylat',
    version='0.1.0a1',
    packages=['pylat', 'pylat.data', 'pylat.util', 'pylat.wrapper',
              'pylat.wrapper.predictor', 'pylat.wrapper.transformer',
              'pylat.neuralnet', 'pylat.neuralnet.rnn'],
    url='https://github.com/alejgh/pylat',
    license='GPL-3.0',
    author='Alejandro González Hevia',
    author_email='alejandrgh11@gmail.com',
    description='A simple library with some common nlp operations',
    install_requires=[
        'gensim', 'scikit-learn', 'numpy', 'pandas',
        'spacy', 'tensorflow==1.13.1', 'xmltodict',
        'en_core_web_sm', 'es_core_news_sm'
    ],
    dependency_links=[
        'https://github.com/explosion/spacy-models/releases/download/en_core_web_sm-2.1.0/en_core_web_sm-2.1.0.tar.gz#egg=en_core_web_sm-2.1.0',
        'https://github.com/explosion/spacy-models/releases/download/es_core_news_sm-2.1.0/es_core_news_sm-2.1.0.tar.gz#egg=es_core_news_sm-2.1.0'
    ]
)



import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin


class AttributesAdder(BaseEstimator, TransformerMixin):
    def __init__(self, transformers):
        """

        :param transformers: List of transform functions that generate
        new features for the model.
        """
        self.transformers = transformers

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        new_features = []
        for transformer in self.transformers:
            new_feature = []
            for sentence in X:
                new_feature.append(transformer(sentence))
            new_features.append(new_feature)
        return np.asarray(new_features).T.astype(float)



# -*- coding: utf-8 -*-

import logging, json, requests, uuid

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class TextTranslator():
    def __init__(self, translator_key):
        self.base_url = 'https://api.cognitive.microsofttranslator.com'
        self.path = '/translate?api-version=3.0'
        self.headers = {
            'Ocp-Apim-Subscription-Key': translator_key,
            'Content-type': 'application/json',
            'X-ClientTraceId': str(uuid.uuid4())
        }

    def translate(self, text, src, to):
        params = '&from={}&to={}'.format(src, to)
        constructed_url = self.base_url + self.path + params
        body = [{
            'text' : text[:5000] # limit maximum number of characters to 5000
        }]
        request = requests.post(constructed_url, json=body, headers=self.headers)
        output = request.json()
        logger.info('Text: {} - Output: {}'.format(text, output))
        return output



import tensorflow as tf

from abc import ABC, abstractmethod
from pylat.exceptions import InvalidArgumentError


class LayerConfig():
    def __init__(self, num_units=None, kernel_init=None, activation=None):
        self.num_units = num_units
        self.kernel_init = kernel_init
        self.activation = activation

    def get_params(self, deep=True):
        return {
            "num_units": self.num_units,
            "kernel_init": self.kernel_init,
            "activation": self.activation,
            "dropout_rate": self.dropout_rate
        }


class BaseLayer(ABC):
    def __init__(self, num_units, kernel_init='glorot_uniform',
                 activation='relu', dropout_rate=None):
        self.num_units = num_units
        self.kernel_init = kernel_init
        self.activation = activation
        self.dropout_rate = dropout_rate

    def load_config(self, layer_config):
        if not isinstance(layer_config, LayerConfig):
            error_msg = "Layer config must be an instance of the LayerConfig" \
                        "class."
            raise InvalidArgumentError(layer_config, error_msg)

        for parameter, value in layer_config.get_params().items():
            if value is not None:
                setattr(self, parameter, value)
        return self

    @abstractmethod
    def build_tensor(self, inputs, **kwargs):
        pass


class DenseLayer(BaseLayer):
    def build_tensor(self, inputs, **kwargs):
        output = tf.keras.layers.Dense(self.num_units,
                                     kernel_initializer=self.kernel_init,
                                     activation=self.activation)(inputs)
        if self.dropout_rate is not None:
            output = tf.keras.layers.Dropout(self.dropout_rate)(output,
                                                                **kwargs)
        return output
